package business;

/**
   This class provides support for loading and playing
   MIDI files from URLs. It provides methods to
   load, play, pause, stop midi play.
   It provides a BoundedRangeModel which controls
   the play position of the current track.
   It provides a BoundedRangeModel which controls
   the volume.
   It allows a ChangeListener to be added to hear
   when a song reaches its end.
 */

import access.DatabaseConstants;

import javax.sound.midi.*;
import javax.sound.sampled.AudioSystem;
import java.util.Vector;
import javax.swing.event.*;
import javax.swing.DefaultBoundedRangeModel;
import javax.swing.BoundedRangeModel;
import javax.swing.Timer;
import java.awt.event.*;
import java.io.File;
import java.io.FileInputStream;
import java.io.BufferedInputStream;
//import java.net.URL;
//import javax.swing.SwingUtilities;

public class Player implements DatabaseConstants, MetaEventListener {
  private Synthesizer synthesizer;
  private Sequencer sequencer;
  private Sequence currentSound;

  // Is the player currently playing
  private boolean playing;
  String sProgramName=null;

  public boolean isPlaying() {
    return playing;
  }

    public void mute(boolean bStatus)
    {
        MidiChannel[] objChannels = synthesizer.getChannels();
         for(int i=0;i<objChannels .length;i++) {
	        objChannels [i].setMute(bStatus);
         }
    }

  // Is the player currently paused
  private boolean paused;
  // Is the player currently holding
  private boolean holding;
  // What position we are holding at
  private int holdPosition;

  public boolean isPaused() {
    return paused;
  }

  public boolean isHolding() {
    return holding;
  }

  private Position position;
  private Gain gain;

    public void setBoundedRangeModel(Position boundedPosition)
    {
         position= boundedPosition;
    }

  /** Return the model for the song position */
  public BoundedRangeModel getPosition() {return position;}

  /** Return the model for the volume */
  public BoundedRangeModel getVolume() {return gain;}

  public Player(String sProgramName) {
    this.sProgramName=sProgramName;

  }

  public boolean getRecordedProgram()
  {
       init();
    if(retrieveRecordedProgram()==null)
        return false;


    position=new Position();
    gain=new Gain();
      //play(sequence);
     return true;
  }

  /** Method to initialise the player. It should:
      1 Get a Sequencer (this may throw an exception
        which we catch).
      2 Add a MetaEventListener to the Sequencer
        to find out when play stops
      3 Get a synthesizer by calling MidiSystem.getSynthesizer()
  */
  private void init() {
    try {
      sequencer = MidiSystem.getSequencer();
      sequencer.addMetaEventListener(this);

      synthesizer = MidiSystem.getSynthesizer();
    }
    catch (MidiUnavailableException e) {
      System.out.println("Can't access MIDI device");
    }
  }

  /** Play a given loaded song */
  public void play() {
    if(sequencer!=null) {
         //if(sequencer!=null && sound!=null && sound instanceof Sequence) {
    //  currentSound=(Sequence) sound;

      // Open the sequencer
      try {
	//sequencer.open();
	synthesizer.open();
      }
      catch (MidiUnavailableException e) {
	System.out.println(e.toString());
	return;
      }

      // Set the current sequence
     /* try {
	sequencer.setSequence((Sequence) currentSound);
      }
      catch (InvalidMidiDataException e) {
	System.out.println(e.toString());
	sequencer.close();
	return;
      }*/
      catch (IllegalArgumentException e) {
	System.out.println(e.toString());
	sequencer.close();
	return;
      }

      // Start the sequencer playing the sound
      sequencer.start();

      // Set up necessary variables
      playing=true;
      ended=false;

      // Tell the position timer to start
      position.startPlay();
    }
  }

  /** Stop the play of the current sound */
  public void stop() {
    if(playing) {
      // Re-initialize variables
      playing=false;
      paused=false;
      holding=false;

      currentSound=null;

      // Tell the player to stop play
      position.stopPlay();

      // Stop the sequencer
      sequencer.stop();

      // Close the synthesizer
      synthesizer.close();

      // Close the sequencer to release the device
      sequencer.close();
    }
  }


  /** Pause or unpause play of playing song. We can only alter the pause status
      if a song is currently playing. If so if the song is paused then resume play
      otherwise pause it. */
  public  void pause() {
    if(playing) {
      if(!paused) {
	paused=true;
	sequencer.stop();
      }
      else {
	paused=false;
	sequencer.start();
      }
    }
  }

  /** Hold or unhold play of playing song. We can only alter the hold status
      if a song is currently playing. If so if the song is holding then resume play
      otherwise loop every time the timer fires (takes effect in the Position
      actionPerformed() method */
  public  void hold() {
    if(playing) {
      if(!holding) {
	holding=true;
        holdPosition = position.getValue();
      }
      else {
	holding=false;
      }
    }
  }

  /** The position class contains a bounded range model which controls the position
      in a given song. It models a value in milliseconds.
      The method getMaximum() returns the length of the current song.
      It returns 0 if there is no current song;

      The method getValue() returns the position in the current song.
      The method setValue() sets the position in the current song.

      It uses a Timer which should run while a song is playing. The
      timer goes off ever 100 milliseconds. Every time it fires
      if the song is playing, it updates the value of the BoundedRangeModel.
  */
  private class Position extends DefaultBoundedRangeModel implements ActionListener {
    /** Set the value. This should:
	1 Set the position of the current track.
	2 Then set the value
    */
    public void setValue(int newValue) {
      if(playing) {
        super.setValue(newValue);
        setPosition(super.getValue());
        if(holding) holdPosition = newValue;
      }
      else{
	super.setValue(0);
      }
    }

    /** Get the current position in the current sequence by
	asking the sequencer
    */
    private int getPosition() {
      return (int)  (sequencer.getMicrosecondPosition()/1000);
    }

    /** Set the current position in the sequencer */
    private void setPosition(int value) {
      sequencer.setMicrosecondPosition(value * 1000);
    }

    /** To create a Position bounded range model, create a Timer */
    public Position() {
      timer=new Timer(100,this);
    }

    /** When a song is playing the bounded range model is updated about the current
	position by a timer */

    private Timer timer;

    private boolean started=false;

    /** Tell the timer that we've started playing a track &
        because the current song will have changed, update the range length */
    public void startPlay() {
      timer.start();
      updLength();

      started=true;
    }

    /** Tell the timer that we've stopped playing a track
        because the current song will have changed, update the range length */
    public void stopPlay() {
      timer.stop();
      updLength();
    }


    /** Update the length of the track. This should be called when a new song
	starts playing. It calculates the length of the song and then sets
	the maximum
    */
    private void updLength() {
      int length;
       
        currentSound=sequencer.getSequence();

      if(currentSound!=null) {
	        length = (int)(((Sequence) currentSound).getMicrosecondLength() / 1000);
      }
      else{
	        length=0;
      }
      super.setMaximum(length);

    }

    /** When the timer fires, if a track is playing and it has not ended, we set
	the value of the BoundedRangeModel with the current position.
	Otherwise we tell the player to stop
    */
    public void actionPerformed(ActionEvent e) {
      if(playing && !ended) {
	// There appears to be some strange behavior whereby the volume
	// can only be set after a song has really started playing.
	// So we make sure that the volume is set correctly the first
	// time a tick fires when a song is playing. (ie 100 milliseconds
	// after the song start). This is important because if we close
	// and then later open the synthesizer the volume changes.

	if(started) {
	  gain.updateGain();started=false;
	}
        if(holding && (getPosition() >= (holdPosition + 1000))) {
          setPosition(holdPosition);
          super.setValue(holdPosition);
        }
        else{
	  super.setValue(getPosition());
            //System.out.println("updated%%%%%%%%%%%%%%%%%%%%%%%%");
        }
      }
      else {
	stop();
      }

      // Tell anything listening that a change has occured.
      onChange();
    }
  }

    public Sequencer retrieveRecordedProgram()
    {
        File fProgram = new File(RECORDED_PROGRAM + sProgramName + RECORDED_PROGRAM_FILE_EXTENSION);
        //String sCurrentName =  fProgram.getName();
        Object currentSound = null;

        try {
            currentSound = AudioSystem.getAudioInputStream(fProgram);

        } catch(Exception e1) {
            // load midi as inputstreams
            try {
                FileInputStream is = new FileInputStream(fProgram);
                currentSound = new BufferedInputStream(is, 1024);
            } catch (Exception e3) {
                //e3.printStackTrace();
                currentSound = null;
                return null;
            }
        }
        if (currentSound instanceof Sequence || currentSound instanceof BufferedInputStream) {
            try {
                sequencer.open();
                if (currentSound instanceof Sequence) {
                    sequencer.setSequence((Sequence) currentSound);

                } else {
                    sequencer.setSequence((BufferedInputStream) currentSound);


                }
            } catch (InvalidMidiDataException imde) {
                System.out.println("Unsupported audio file.");
                currentSound = null;
                return null;
            } catch (Exception ex) {
                ex.printStackTrace();
                currentSound = null;
                return null;
            }
        }
        //this.sequencer=sequencer;
        return sequencer;
    }


  /** When a track ends it sends a MetaMessage of type 47.
      When this happens simply set the boolean value "ended" to true to tell
      the Player that the song has ended.
   */
  private boolean ended=false;
  public void meta(MetaMessage message) {
    if (message.getType() == 47) {  // 47 is end of track
      ended=true;
    }
  }

  /** The Gain object controls and models the volume. It models it as a percentage.
      So values are between 0 & 100. By default the volume of the synthesizer is
      at 80% */
  private class Gain extends DefaultBoundedRangeModel {
    public Gain() {
      super(80,0,0,100);
    }
    /* To set the value. Set the actual volume and then set the value of the model
     */
    public void setValue(int pos) {
      super.setValue(pos);
      updateGain();
    }

    /** Set the volume. To do this
	1 Convert the percentage value into an int value between 0 & 127.
	2 Get an array of all the midi channels
	3 Update the volume by changin the volume of each midi channel
    */
    private void updateGain() {
      int pos=getValue();
      int value = (int) (127 * (double) (pos / 100.0));
      MidiChannel[] cs = synthesizer.getChannels();
      for(int i=0;i<cs.length;i++) {
	        cs[i].controlChange(7,value);

      }
    }
  }



  /** Support change Listeners that may be added & removed.
      When a song reaches the end and stops a ChangeEvent is fired to all
      listeners */
  public void addChangeListener(ChangeListener l) {
    listeners.add(l);
  }

  public void removeChangeListener(ChangeListener l) {
    listeners.remove(l);
  }

  /** When a song reaches the end, generate a ChangeEvent with the player as
      its source.
      Then get an array from the listeners vector and loop through firing
      each listener
  */
  public void onChange() {
    ChangeEvent e = new ChangeEvent(this);
    Object[] ls = listeners.toArray();
    for(int i=0;i<ls.length;i++) {

      ((ChangeListener) ls[i]).stateChanged(e);
    }
  }

  private Vector listeners=new Vector();
}
